/*
 * deviced
 *
 * Copyright (c) 2012 - 2013 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <assert.h>
#include <sys/time.h>
#include <fcntl.h>
#include <signal.h>
#include <dirent.h>
#include <ctype.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <poll.h>
#include <mntent.h>
#include <system_info.h>
#include <sys/mount.h>
#include "log.h"
#include "common.h"

#define PERMANENT_DIR   "/tmp/permanent"
#define VIP_DIR         "/tmp/vip"
#define BUFF_MAX        255

#define APP_ATTR_PATH "/proc/%d/attr/current"

/**
 * Opens "/proc/$pid/oom_score_adj" file for w/r;
 * Return: FILE pointer or NULL
 */
FILE *open_proc_oom_score_adj_file(int pid, const char *mode)
{
	char buf[32];
	FILE *fp;

	snprintf(buf, sizeof(buf), "/proc/%d/oom_score_adj", pid);
	fp = fopen(buf, mode);
	return fp;
}

int get_cmdline_name(pid_t pid, char *cmdline, size_t cmdline_size)
{
	int fd, ret;
	char buf[PATH_MAX + 1];
	char *filename;

	snprintf(buf, sizeof(buf), "/proc/%d/cmdline", pid);
	fd = open(buf, O_RDONLY);
	if (fd < 0) {
		errno = ESRCH;
		return -1;
	}

	ret = read(fd, buf, PATH_MAX);
	close(fd);
	if (ret < 0)
		return -1;

	buf[PATH_MAX] = '\0';

	filename = strrchr(buf, '/');
	if (filename == NULL)
		filename = buf;
	else
		filename = filename + 1;

	if (cmdline_size < strlen(filename) + 1) {
		errno = EOVERFLOW;
		return -1;
	}

	strncpy(cmdline, filename, cmdline_size - 1);
	cmdline[cmdline_size - 1] = '\0';
	return 0;
}

int is_vip(int pid)
{
	if (pid < 1)
		return -1;

	char buf[PATH_MAX];

	snprintf(buf, PATH_MAX, "%s/%d", VIP_DIR, pid);

	if (access(buf, R_OK) == 0)
		return 1;
	else
		return 0;
}

static int remove_dir_internal(int fd)
{
	DIR *dir;
	struct dirent *de = NULL;
	int subfd, ret = 0;

	dir = fdopendir(fd);
	if (!dir)
		return -1;
	while (1) {
		de = readdir(dir);
		if (ret != 0 || de == NULL)
			break;
		if (de->d_type == DT_DIR) {
			if (!strncmp(de->d_name, ".", 2) || !strncmp(de->d_name, "..", 3))
				continue;
			subfd = openat(fd, de->d_name, O_RDONLY | O_DIRECTORY);
			if (subfd < 0) {
				_SE("Couldn't openat %s: %d\n", de->d_name, errno);
				ret = -1;
				continue;
			}
			if (remove_dir_internal(subfd))
				ret = -1;

			close(subfd);
			if (unlinkat(fd, de->d_name, AT_REMOVEDIR) < 0) {
				_SE("Couldn't unlinkat %s: %d\n", de->d_name, errno);
				ret = -1;
			}
		} else {
			if (unlinkat(fd, de->d_name, 0) < 0) {
				_SE("Couldn't unlinkat %s: %d\n", de->d_name, errno);
				ret = -1;
			}
		}
	}
	closedir(dir);
	return ret;
}

int remove_dir(const char *path, int del_dir)
{
	int fd, ret = 0;

	if (!path)
		return -1;
	fd = open(path, O_RDONLY | O_NONBLOCK | O_DIRECTORY | O_CLOEXEC | O_NOFOLLOW);
	if (fd < 0) {
		_SE("Couldn't opendir %s: %d\n", path, errno);
		return -errno;
	}
	ret = remove_dir_internal(fd);
	close(fd);

	if (del_dir) {
		if (rmdir(path)) {
			_SE("Couldn't rmdir %s: %d\n", path, errno);
			ret = -1;
		}
	}
	return ret;
}

/*
 * Helper function
 * - Read from sysfs entry
 * - Write to sysfs entry
 */
int sys_check_node(char *path)
{
	int fd;

	fd = open(path, O_RDONLY);
	if (fd == -1)
		return -1;

	close(fd);
	return 0;
}

static int sys_read_buf(char *file, char *buf, int len)
{
	int fd;
	int r;
	int ret = 0;

	fd = open(file, O_RDONLY);
	if (fd == -1)
		return -ENOENT;

	r = read(fd, buf, len);
	if ((r >= 0) && (r < len))
		buf[r] = '\0';
	else {
		buf[0] = '\0';
		ret = -EIO;
	}

	close(fd);

	return ret;
}

static int sys_write_buf(char *file, char *buf)
{
	int fd;
	int r;
	int ret = 0;

	fd = open(file, O_WRONLY);
	if (fd == -1)
		return -ENOENT;

	r = write(fd, buf, strlen(buf));
	if (r < 0)
		ret = -EIO;

	close(fd);

	return ret;
}

int sys_get_int(char *fname, int *val)
{
	char buf[BUFF_MAX];
	int ret = 0;

	if (!fname || !val)
		return -EINVAL;

	if (sys_read_buf(fname, buf, sizeof(buf)) == 0) {
		*val = atoi(buf);
	} else {
		*val = -1;
		ret = -EIO;
	}

	return ret;
}

int sys_set_int(char *fname, int val)
{
	char buf[BUFF_MAX];
	int ret = 0;

	snprintf(buf, sizeof(buf), "%d", val);

	if (sys_write_buf(fname, buf) != 0)
		ret = -EIO;

	return ret;
}

int sys_get_str(char *fname, char *str, int len)
{
	int ret;

	if (!fname || !str || len < 0)
		return -EINVAL;

	ret = sys_read_buf(fname, str, len);
	if (ret < 0)
		return ret;
	return 0;
}

int sys_set_str(char *fname, char *val)
{
	int r = -1;

	if (val != NULL) {
		if (sys_write_buf(fname, val) == 0)
			r = 0;
	}

	return r;
}

static int terminate_processes_on_partition(const char *partition, bool force)
{
	const char *argv[7] = {"/usr/bin/fuser", "-m", "-k", "-s", NULL, NULL, NULL};
	int argc;

	if (force)
		argv[4] = "-SIGKILL";
	else
		argv[4] = "-SIGTERM";

	argv[5] = partition;
	argc = sizeof(argv) / sizeof(argv[0]);

	return run_child(argc, argv);
}

int mount_check(const char *path)
{
	int ret = false;
	struct mntent *mnt;
	const char *table = "/etc/mtab";
	FILE *fp;
	int len;

	fp = setmntent(table, "r");
	if (!fp)
		return ret;

	len = strlen(path) + 1;
	while (1) {
		mnt = getmntent(fp);
		if (mnt == NULL)
			break;
		if (!strncmp(mnt->mnt_dir, path, len)) {
			ret = true;
			break;
		}
	}
	endmntent(fp);
	return ret;
}

static int get_num_processes_on_partition(const char *part_path)
{
	FILE *fp;
	char *cmd = NULL;
	char *line = NULL;
	size_t len = 0;
	int num_processes = 0;
	int ret = 0;

	ret = asprintf(&cmd, "fuser -m %s | grep -o '[0-9]*'", part_path);
	if (ret < 0)
		return -1;

	printf("cmd=%s\n", cmd);
	fp = popen(cmd, "r");
	free(cmd);
	if (fp == NULL)
		return -1;

	while (getline(&line, &len, fp) != -1)
		num_processes++;

	free(line);
	pclose(fp);

	return num_processes;
}

void umount_partition_by_kill(const char *path, const int max_retry)
{
	FILE *fp;
	char *part_path = NULL;
	size_t len = 0;
	int ret = 0, retry = 0;
	int remain;
	char *cmd = NULL;

	/* if the path is not a mountpoint, do not perform umounting and killing */
	if (!mount_check(path))
		return;

	ret = asprintf(&cmd, "mount | grep \" on %s \" | awk '{print $1}'", path);
	if (ret < 0)
		return;

	fp = popen(cmd, "r");
	free(cmd);
	if (!fp)
		return;

	ret = getline(&part_path, &len, fp);
	if (ret == -1 || !part_path) {
		pclose(fp);
		free(part_path);
		return;
	} else if (ret > 0 && *(part_path + ret - 1) == '\n') {
		*(part_path + ret -1) = '\0';
	}

	sync();
	umount2(path, MNT_DETACH);

	do {
		/* Kill processes with SIGTERM */
		terminate_processes_on_partition(part_path, false);
		usleep((useconds_t)MSEC_TO_USEC(500));

		/* Kill processes with SIGKILL */
		terminate_processes_on_partition(part_path, true);
		usleep((useconds_t)MSEC_TO_USEC(200));

		remain = get_num_processes_on_partition(part_path);

		retry++;

	} while (remain > 0 && retry < max_retry);
	sync();

	free(part_path);
	pclose(fp);
	return;
}

#define CMD_FROZEN "FROZEN"
#define CMD_THAWED "THAWED"
static int freeze_processes_on_path(const char *frz_name, const char *part_path)
{
	char *fuser_cmd = NULL;
	FILE *fuser_fp;
	char *line = NULL;
	size_t len = 0;
	int task_fd;
	int freezer_fd;
	char *freezer_procs;
	char *freezer_state;
	int ret;

	ret = asprintf(&fuser_cmd, "fuser -m %s | grep -o '[0-9]*'", part_path);
	if (ret < 0)
		return -1;

	fuser_fp = popen(fuser_cmd, "r");
	free(fuser_cmd);
	if (fuser_fp == NULL)
		return -1;

	ret = asprintf(&freezer_procs, "/sys/fs/cgroup/freezer/%s/cgroup.procs", frz_name);
	if (ret < 0) {
		pclose(fuser_fp);
		return -1;
	}

	task_fd = open(freezer_procs, O_RDWR);
	free(freezer_procs);
	if (task_fd < 0) {
		pclose(fuser_fp);
		return -1;
	}

	while (getline(&line, &len, fuser_fp) != -1) {
		int ret;
		int str_len = strlen(line);
		if (str_len <= 1)
			continue;

		ret = write(task_fd, line, str_len-1);
		if (ret < 0)
			break;
	}

	close(task_fd);
	free(line);
	pclose(fuser_fp);

	ret = asprintf(&freezer_state, "/sys/fs/cgroup/freezer/%s/freezer.state", frz_name);
	if (ret < 0)
		return -1;

	freezer_fd = open(freezer_state, O_RDWR);
	free(freezer_state);
	if (freezer_fd < 0)
		return -1;

	ret = write(freezer_fd, CMD_FROZEN, strlen(CMD_FROZEN));
	close(freezer_fd);
	if (ret < 0)
		return -1;

	return 0;
}

static bool check_frozen(const char *frz_name)
{
	char *state_path = NULL;
	char *line = NULL;
	size_t len = 0;
	FILE *fd;
	int ret;
	bool result = false;

	ret = asprintf(&state_path, "/sys/fs/cgroup/freezer/%s/freezer.state", frz_name);
	if (ret < 0)
		return false;

	fd = fopen(state_path, "r");
	free(state_path);
	if (fd == NULL)
		return false;

	ret = getline(&line, &len, fd);
	if (ret != -1 && line) {
		if (!strncmp(line, CMD_FROZEN, strlen(CMD_FROZEN)))
			result = true;
		else
			result = false;

	}

	free(line);
	fclose(fd);

	return result;
}

static int thaw_processes_on_path(const char *frz_name)
{
	int freezer_fd;
	int ret;

	char *freezer_state;
	ret = asprintf(&freezer_state, "/sys/fs/cgroup/freezer/%s/freezer.state", frz_name);
	if (ret < 0)
		return -1;

	freezer_fd = open(freezer_state, O_RDWR);
	free(freezer_state);
	if (freezer_fd < 0)
		return -1;

	ret = write(freezer_fd, CMD_THAWED, strlen(CMD_THAWED));
	close(freezer_fd);
	if (ret < 0)
		return -1;

	return 0;
}

void suspend_path(const char *frz_name, const char *path, const int max_retry)
{
	bool frozen = false;
	int retry = 0;

	freeze_processes_on_path(frz_name, path);

	do {
		usleep((useconds_t)MSEC_TO_USEC(500));

		frozen = check_frozen(frz_name);

		retry++;

	} while (!frozen && retry < max_retry);
}

void resume_path(const char *frz_name, const char *path)
{
	thaw_processes_on_path(frz_name);
}

void print_time(const char *prefix)
{
	struct timeval tv;
	struct tm tm;
	struct tm *ret;
	gettimeofday(&tv, NULL);
	ret = localtime_r(&(tv.tv_sec), &tm);
	if (ret)
		_D("'%s' %d:%02d:%02d %06ld",
				prefix, tm.tm_hour, tm.tm_min, tm.tm_sec, tv.tv_usec);
}

int get_privilege(pid_t pid, char *name, size_t len)
{
	char path[PATH_MAX];
	char attr[BUFF_MAX];
	size_t attr_len;
	FILE *fp;

	snprintf(path, sizeof(path), APP_ATTR_PATH, pid);

	fp = fopen(path, "r");
	if (!fp)
		return -errno;

	attr_len = fread(attr, 1, sizeof(attr) - 1, fp);
	fclose(fp);
	if (attr_len <= 0)
		return -ENOENT;

	attr[attr_len] = '\0';

	snprintf(name, len, "%s", attr);
	return 0;
}

#define MODEL_NAME      "http://tizen.org/system/model_name"
#define MODEL_EMULATOR  "Emulator"

bool is_emulator(void)
{
	int ret;
	char *model_name = NULL;
	static bool emul = false;
	static int set = 0;

	if (set)
		return emul;

	ret = system_info_get_platform_string(MODEL_NAME, &model_name);
	if (ret < 0) {
		_E("Cannot get model name: %d", ret);
		return emul;
	}

	if (!strncmp(MODEL_EMULATOR, model_name, strlen(model_name) + 1))
		emul = true;

	set = 1;
	free(model_name);

	return emul;
}

int do_mkdir(const char *path, mode_t mode)
{
	char d[PATH_MAX];
	size_t s, l;
	int r, p;

	assert(path);

	l = strlen(path);

	for (p = 0, s = 0; p < l; p += s + 1) {
		s = strcspn(path + p, "/");
		if (!s)
			continue;

		assert(PATH_MAX > p + s + 1);

		r = snprintf(d, p + s + 1, "%s", path);
		if (r < 0)
			return r;

		r = mkdir(d, mode);
		if (r < 0 && errno != EEXIST)
			return -errno;
	}

	return 0;
}

static int do_copy_internal(const char *src, const char *dst, mode_t mode, bool force)
{
	_cleanup_close_ int rfd = -1, wfd = -1;
	char buf[1024];
	ssize_t red;
	int r;

	assert(src);
	assert(dst);

	if (!force) {
		r = access(dst, F_OK);
		if (r == 0)
			return -EALREADY;
		else if (errno != ENOENT)
			return -errno;
	}

	wfd = open(dst, O_CREAT | O_WRONLY | O_TRUNC, mode);
	if (wfd < 0)
		return -errno;

	rfd = open(src, O_RDONLY);
	if (rfd < 0)
		return -errno;

	while ((red = read(rfd, buf, 1024)) > 0)
		if (write(wfd, buf, red) != red)
			return -errno;

	if (red < 0)
		return -errno;

	return 0;
}

int do_copy_force(const char *src, const char *dst)
{
	assert(src);
	assert(dst);

	return do_copy_internal(src, dst, 0644, true);
}

#if 0
int print_open_files(const char *mount_point)
{
	DIR *dp;
	struct dirent entry;
	struct dirent *dentry;

	DIR *dp_child;
	struct dirent entry_child;
	struct dirent *dentry_child;

	int pid = -1, fd;
	int ret;
	char buf[PATH_MAX];
	char buf2[PATH_MAX];

	char cmdline[PATH_MAX];
	char check_path[PATH_MAX];

	int len = strlen(mount_point);

	dp = opendir("/proc");
	if (!dp) {
		_E("FAIL: open /proc");
		return -1;
	}

	while (1) {
		ret = readdir_r(dp, &entry, &dentry);
		if (ret != 0 || dentry == NULL)
			break;

		if (!isdigit(dentry->d_name[0]))
			continue;

		pid = atoi(dentry->d_name);
		snprintf(buf, PATH_MAX, "/proc/%d/cmdline", pid);

		fd = open(buf, O_RDONLY);
		if (fd < 0)
			continue;
		ret = read(fd, cmdline, PATH_MAX);
		close(fd);

		if (ret < 0 || ret >= PATH_MAX)
			continue;
		cmdline[ret] = '\0';

		snprintf(buf, PATH_MAX, "/proc/%d/fd", pid);
		dp_child = opendir(buf);
		if (!dp_child)
			continue;
		while (1) {
			ret = readdir_r(dp_child, &entry_child, &dentry_child);
			if (ret != 0 || dentry_child == NULL)
				break;

			snprintf(check_path, PATH_MAX, "%s/%s", buf, dentry_child->d_name);

			if (readlink(check_path, buf2, PATH_MAX) < 0)
				continue;

			if (strncmp(buf2, mount_point, len) == 0)
				_D("Process %s : Opened files - %s", cmdline, buf2);
		}
		closedir(dp_child);
	}

	closedir(dp);
	return 0;
}

int get_exec_pid(const char *execpath)
{
	DIR *dp;
	struct dirent entry;
	struct dirent *dentry;
	int pid = -1, fd;
	int ret;
	int len;
	char buf[PATH_MAX];
	char buf2[PATH_MAX];

	dp = opendir("/proc");
	if (!dp) {
		_E("FAIL: open /proc");
		return -1;
	}

	len = strlen(execpath) + 1;
	while (1) {
		ret = readdir_r(dp, &entry, &dentry);
		if (ret != 0 || dentry == NULL)
			break;

		if (!isdigit(dentry->d_name[0]))
			continue;

		pid = atoi(dentry->d_name);

		snprintf(buf, PATH_MAX, "/proc/%d/cmdline", pid);
		fd = open(buf, O_RDONLY);
		if (fd < 0)
			continue;
		ret = read(fd, buf2, PATH_MAX);
		close(fd);

		if (ret < 0 || ret >= PATH_MAX)
			continue;

		buf2[ret] = '\0';

		if (!strncmp(buf2, execpath, len)) {
			closedir(dp);
			return pid;
		}
	}

	errno = ESRCH;
	closedir(dp);
	return -1;
}

static int get_num_thawed_processes(const char *path)
{
	char *fuser_cmd = NULL;
	FILE *fuser_fp;
	char *line = NULL;
	size_t len = 0;
	int num_processes = 0;
	int ret;

	ret = asprintf(&fuser_cmd, "fuser -m %s | grep -o '[0-9]*'", path);
	if (ret < 0)
		return -1;

	fuser_fp = popen(fuser_cmd, "r");
	free(fuser_cmd);
	if (fuser_fp == NULL)
		return -1;

	while (getline(&line, &len, fuser_fp) != -1) {
		char *stat_cmd = NULL;
		FILE *stat_fp;
		int pid;
		char comm[128];
		char state = 0;
		int str_len;
		int ret;

		str_len = strlen(line);
		if (str_len > 0)
			// To remove '\n'
			*(line+str_len-1) = 0;

		ret = asprintf(&stat_cmd, "cat /proc/%s/stat", line);
		if (ret < 0) {
			num_processes++;
			continue;
		}

		stat_fp = popen(stat_cmd, "r");
		free(stat_cmd);
		if (stat_fp == NULL) {
			num_processes++;
			continue;
		}

		fscanf(stat_fp, "%d %128s %c", &pid, comm, &state);
		pclose(stat_fp);

		// if thawed
		if (state != 'D') num_processes++;
	}

	free(line);
	pclose(fuser_fp);

	return num_processes;
}
#endif

