/*
 * deviced
 *
 * Copyright (c) 2012 - 2013 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


/**
 * @file	display-dbus.c
 * @brief	dbus interface
 *
 */

#include <error.h>
#include <stdbool.h>
#include <device-node.h>
#include <libsyscommon/dbus-system.h>
#include <device/display.h>

#include "ambient-mode.h"
#include "core/log.h"
#include "util.h"
#include "core.h"
#include "lock-detector.h"
#include "core/common.h"
#include "core/devices.h"
#include "core/device-idler.h"
#include "core/device-notifier.h"
#include "apps/apps.h"
#include "dd-display.h"
#include "display-actor.h"
#include "display-ops.h"

#define SIGNAL_HOMESCREEN       "HomeScreen"

#define AUL_APPSTATUS_PATH		"/Org/Tizen/Aul/AppStatus"
#define AUL_APPSTATUS_INTERFACE		"org.tizen.aul.AppStatus"
#define APP_CHANGE_STATE		"AppStatusChange"

#define TELEPHONY_PATH			"/org/tizen/telephony"
#define TELEPHONY_INTERFACE_SIM		"org.tizen.telephony.Manager"
#define SIGNAL_SIM_STATUS		"SimInserted"
#define SIM_CARD_NOT_PRESENT		(0x0)

#ifndef VCONFKEY_LCD_BRIGHTNESS_INIT
#define VCONFKEY_LCD_BRIGHTNESS_INIT	"db/private/deviced/lcd_brightness_init"
#endif

#define DISPLAY_DIM_BRIGHTNESS  0
#define DUMP_MODE_WATING_TIME   600000

#define LCDOFF_PROXI_STR		"proximity"
#define LCDOFF_GESTURE_STR		"gesture"

#define EXPIRED_POPUP_TYPE_POWER   "power_lock_expired"
#define EXPIRED_POPUP_PID    "_APP_PID_"
#define EXPIRED_POPUP_COMM   "_APP_COMM_"
#define EXPIRED_POPUP_ID     "_REQUEST_ID_"

GVariant *dbus_start(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	static const struct device_ops *display_device_ops = NULL;

	if (!display_device_ops)
		display_device_ops = find_device("display");
	if (NOT_SUPPORT_OPS(display_device_ops))
		goto out;

	display_device_ops->start(CORE_LOGIC_MODE);
out:
	return dbus_handle_new_g_variant_tuple();
}

GVariant *dbus_stop(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	static const struct device_ops *display_device_ops = NULL;

	if (!display_device_ops)
		display_device_ops = find_device("display");
	if (NOT_SUPPORT_OPS(display_device_ops))
		goto out;

	display_device_ops->stop(CORE_LOGIC_MODE);
out:
	return dbus_handle_new_g_variant_tuple();
}

GVariant *dbus_lockstate(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	char *state_str;
	char *option1_str;
	char *option2_str;
	int timeout;
	pid_t pid;
	int state;
	int flag;
	int ret = 0;
	unsigned int caps;

	g_variant_get(param, "(sssi)", &state_str, &option1_str, &option2_str, &timeout);

	if (!state_str || timeout < 0) {
		_E("message is invalid!");
		ret = -EINVAL;
		goto out;
	}

	if (!strcmp(state_str, "privilege check"))
		goto out;

	pid = dbus_connection_get_sender_pid(conn, sender);
	if (kill(pid, 0) == -1) {
		_E("%d process does not exist, dbus ignored!", pid);
		ret = -ESRCH;
		goto out;
	}

	if (!strcmp(state_str, PM_LCDON_STR))
		state = LCD_NORMAL;
	else if (!strcmp(state_str, PM_LCDDIM_STR))
		state = LCD_DIM;
	else if (!strcmp(state_str, PM_LCDOFF_STR))
		state = LCD_OFF;
	else {
		_E("%s state is invalid, dbus ignored.", state_str);
		ret = -EINVAL;
		goto out;
	}

	if (!strcmp(option1_str, STAYCURSTATE_STR))
		flag = STAY_CUR_STATE;
	else if (!strcmp(option1_str, GOTOSTATENOW_STR))
		flag = GOTO_STATE_NOW;
	else {
		_E("%s option is invalid. set default option!", option1_str);
		flag = STAY_CUR_STATE;
	}

	if (!strcmp(option2_str, HOLDKEYBLOCK_STR))
		flag |= HOLD_KEY_BLOCK;

	if (flag & GOTO_STATE_NOW) {
		caps = display_get_caps(DISPLAY_ACTOR_API);

		if (!display_has_caps(caps, DISPLAY_CAPA_LCDON) &&
		    state == LCD_NORMAL) {
			_D("No lcdon capability!");
			ret = -EPERM;
			goto out;
		}
		if (!display_has_caps(caps, DISPLAY_CAPA_LCDOFF) &&
		    state == LCD_OFF) {
			_D("No lcdoff capability!");
			ret = -EPERM;
			goto out;
		}
	}

	if (check_dimstay(state, flag) == true) {
		_E("LCD state can not be changed to OFF state now!");
		flag &= ~GOTO_STATE_NOW;
		flag |= STAY_CUR_STATE;
	}

	if (disp_plgn.pm_lock_internal)
		ret = disp_plgn.pm_lock_internal(pid, state, flag, timeout);
out:
	g_free(state_str);
	g_free(option1_str);
	g_free(option2_str);
	return g_variant_new("(i)", ret);
}

GVariant *dbus_unlockstate(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	char *state_str;
	char *option_str;
	pid_t pid;
	int state;
	int flag;
	int ret = 0;

	g_variant_get(param, "(ss)", &state_str, &option_str);

	if (!state_str) {
		_E("message is invalid!");
		ret = -EINVAL;
		goto out;
	}

	if (!strcmp(state_str, "privilege check"))
		goto out;

	pid = dbus_connection_get_sender_pid(conn, sender);
	if (kill(pid, 0) == -1) {
		_E("%d process does not exist, dbus ignored!", pid);
		ret = -ESRCH;
		goto out;
	}

	if (!strcmp(state_str, PM_LCDON_STR))
		state = LCD_NORMAL;
	else if (!strcmp(state_str, PM_LCDOFF_STR))
		state = LCD_OFF;
	else {
		_E("%s state is invalid, dbus ignored!", state_str);
		ret = -EINVAL;
		goto out;
	}

	if (!strcmp(option_str, SLEEP_MARGIN_STR))
		flag = PM_SLEEP_MARGIN;
	else if (!strcmp(option_str, RESET_TIMER_STR))
		flag = PM_RESET_TIMER;
	else if (!strcmp(option_str, KEEP_TIMER_STR))
		flag = PM_KEEP_TIMER;
	else {
		_E("%s option is invalid. set default option!", option_str);
		flag = PM_RESET_TIMER;
	}

	if (disp_plgn.pm_unlock_internal)
		ret = disp_plgn.pm_unlock_internal(pid, state, flag);
out:
	g_free(state_str);
	g_free(option_str);
	return g_variant_new("(i)", ret);
}

GVariant *dbus_changestate(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	char *state_str;
	pid_t pid;
	int state;
	int ret = 0, len;
	unsigned int caps;

	g_variant_get(param, "(s)", &state_str);

	if (!state_str) {
		_E("message is invalid!");
		ret = -EINVAL;
		goto out;
	}

	if (!strcmp(state_str, "privilege check"))
		goto out;

	pid = dbus_connection_get_sender_pid(conn, sender);
	if (kill(pid, 0) == -1) {
		_E("%d process does not exist, dbus ignored!", pid);
		ret = -ESRCH;
		goto out;
	}

	len = strlen(state_str);
	if (len == 0) {
		ret = -EINVAL;
		goto out;
	}

	if (!strcmp(state_str, PM_LCDON_STR))
		state = LCD_NORMAL;
	else if (!strcmp(state_str, PM_LCDDIM_STR))
		state = LCD_DIM;
	else if (!strcmp(state_str, PM_LCDOFF_STR))
		state = LCD_OFF;
	else if (!strcmp(state_str, PM_STANDBY_STR))
		state = STANDBY;
	else if (!strcmp(state_str, PM_SUSPEND_STR))
		state = SUSPEND;
	else {
		_E("%s state is invalid, dbus ignored!", state_str);
		ret = -EINVAL;
		goto out;
	}

	caps = display_get_caps(DISPLAY_ACTOR_API);

	if (!display_has_caps(caps, DISPLAY_CAPA_LCDON) &&
	    state == LCD_NORMAL) {
		_D("No lcdon capability!");
		ret = -EPERM;
		goto out;
	}
	if (!display_has_caps(caps, DISPLAY_CAPA_LCDOFF) &&
	    state == LCD_OFF) {
		_D("No lcdoff capability!");
		ret = -EPERM;
		goto out;
	}

	if (check_dimstay(state, GOTO_STATE_NOW) == true) {
		_E("LCD state can not be changed to OFF state! by %d", pid);
		ret = -EBUSY;
		goto out;
	}

	if (disp_plgn.pm_change_internal)
		ret = disp_plgn.pm_change_internal(pid, state);

	if (!ret && state == LCD_OFF)
		update_lcdoff_source(VCONFKEY_PM_LCDOFF_BY_TIMEOUT);
out:
	g_free(state_str);
	return g_variant_new("(i)", ret);
}

GVariant *dbus_getdisplaycount(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret;

	ret = DEFAULT_DISPLAY_COUNT;

	return g_variant_new("(i)", ret);
}

GVariant *dbus_getmaxbrightness(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret, state;

	g_variant_get(param, "(i)", &state);
	if (state == DISPLAY_STATE_NORMAL)
		ret = DEFAULT_DISPLAY_MAX_BRIGHTNESS;
	else if (state == DISPLAY_STATE_SCREEN_DIM)
		ret = DEFAULT_DISPLAY_MAX_DIM_BRIGHTNESS;
	else
		ret = -EINVAL;

	return g_variant_new("(i)", ret);
}

GVariant *dbus_setmaxbrightness(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret;

	ret = -ENOTSUP;

	return g_variant_new("(i)", ret);
}

GVariant *dbus_getbrightness(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int brt = -1, state, ret;

	g_variant_get(param, "(i)", &state);

	if (state == DISPLAY_STATE_NORMAL)
		ret = backlight_ops.get_brightness(&brt);
	else if (state == DISPLAY_STATE_SCREEN_DIM) {
		ret = vconf_get_int(VCONFKEY_SETAPPL_LCD_DIM_BRIGHTNESS, &brt);
		if (ret < 0)
			_E("Failed to get vconf value for lcd dim brightness: %d", vconf_get_ext_errno());
	} else {
		ret = backlight_ops.get_default_brt();
		brt = 0;
	}

	if (ret >= 0)
		ret = brt;

	_I("get brightness %d, %d", brt, ret);

	return g_variant_new("(i)", ret);
}

GVariant *dbus_setbrightness(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int state, brt, autobrt, ret = 0, caps;

	caps = display_get_caps(DISPLAY_ACTOR_API);

	if (!display_has_caps(caps, DISPLAY_CAPA_BRIGHTNESS)) {
		_D("No brightness changing capability!");
		ret = -EPERM;
		goto error;
	}

	g_variant_get(param, "(ii)", &state, &brt);

	//Check if brt is same with DIM
	if (brt == DISPLAY_DIM_BRIGHTNESS) {
		_E("application can not set this value(DIM VALUE:%d)", brt);
		ret = -EPERM;
		goto error;
	}

	ret = vconf_get_int(VCONFKEY_SETAPPL_BRIGHTNESS_AUTOMATIC_INT, &autobrt);
	if (ret < 0) {
		_E("Failed to get vconf value for automatic brightness: %d", vconf_get_ext_errno());
		autobrt = SETTING_BRIGHTNESS_AUTOMATIC_OFF;
	}

	if (autobrt == SETTING_BRIGHTNESS_AUTOMATIC_ON) {
		_E("auto_brightness state is ON, can not change the brightness value");
		ret = -EPERM;
		goto error;
	}

	pm_status_flag &= ~DIM_MASK;

	backlight_ops.set_default_brt(brt);
	if (state == DISPLAY_STATE_NORMAL) {
		ret = backlight_ops.set_brightness(brt);
		if (ret < 0)
			goto error;

		ret = vconf_set_int(VCONFKEY_SETAPPL_LCD_BRIGHTNESS, brt);
		if (ret < 0)
			_E("Failed to set vconf value for lcd brightness: %d", vconf_get_ext_errno());
	} else if (state == DISPLAY_STATE_SCREEN_DIM) {
		if (pm_cur_state == S_LCDDIM) {
			ret = backlight_ops.set_brightness(brt);
			if (ret < 0)
				goto error;
		}

		ret = vconf_set_int(VCONFKEY_SETAPPL_LCD_DIM_BRIGHTNESS, brt);
		if (ret < 0)
			_E("Failed to set vconf value for lcd dim brightness: %d", vconf_get_ext_errno());
	} else
		ret = -EINVAL;

	_I("set brightness %d, %d", brt, ret);

error:
	return g_variant_new("(i)", ret);
}

GVariant *dbus_holdbrightness(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int brt, autobrt, ret, caps;

	caps = display_get_caps(DISPLAY_ACTOR_API);

	if (!display_has_caps(caps, DISPLAY_CAPA_BRIGHTNESS)) {
		_D("No brightness changing capability!");
		ret = -EPERM;
		goto error;
	}

	g_variant_get(param, "(i)", &brt);

	if (brt == DISPLAY_DIM_BRIGHTNESS) {
		_E("application can not set this value(DIM VALUE:%d)", brt);
		ret = -EPERM;
		goto error;
	}

	ret = vconf_get_int(VCONFKEY_SETAPPL_BRIGHTNESS_AUTOMATIC_INT, &autobrt);
	if (ret < 0) {
		_E("Failed to get vconf value for automatic brightness: %d", vconf_get_ext_errno());
		autobrt = SETTING_BRIGHTNESS_AUTOMATIC_OFF;
	}

	ret = vconf_set_int(VCONFKEY_PM_CUSTOM_BRIGHTNESS_STATUS, VCONFKEY_PM_CUSTOM_BRIGHTNESS_ON);
	if (ret < 0)
		_E("Failed to set vconf value for custom brightness status: %d", vconf_get_ext_errno());

	ret = backlight_ops.set_brightness(brt);
	if (ret < 0)
		goto error;

	if (autobrt == SETTING_BRIGHTNESS_AUTOMATIC_ON) {
		_D("Auto brightness will be paused");
		ret = vconf_set_int(VCONFKEY_SETAPPL_BRIGHTNESS_AUTOMATIC_INT, SETTING_BRIGHTNESS_AUTOMATIC_PAUSE);
		if (ret < 0)
			_E("Failed to set vconf value for automatic brightness: %d", vconf_get_ext_errno());
	}

	_I("hold brightness %d, %d", brt, ret);

error:
	return g_variant_new("(i)", ret);
}

GVariant *dbus_releasebrightness(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int bat, charger, changed, setting, brt, autobrt, ret = 0;

	ret = vconf_get_int(VCONFKEY_SYSMAN_BATTERY_STATUS_LOW, &bat);
	if (ret < 0) {
		_E("Failed to get vconf value for battery status low: %d", vconf_get_ext_errno());
		ret = -EPERM;
		goto error;
	}

	ret = vconf_get_int(VCONFKEY_SYSMAN_CHARGER_STATUS, &charger);
	if (ret < 0) {
		_E("Failed to get vconf value for charger status: %d", vconf_get_ext_errno());
		ret = -EPERM;
		goto error;
	}

	ret = vconf_get_bool(VCONFKEY_PM_BRIGHTNESS_CHANGED_IN_LPM, &changed);
	if (ret < 0) {
		_E("Failed to get vconf value for brightness changed in lpm: %d", vconf_get_ext_errno());
		ret = -EPERM;
		goto error;
	}

	ret = vconf_get_int(VCONFKEY_SETAPPL_LCD_BRIGHTNESS, &setting);
	if (ret < 0) {
		_E("Failed to get vconf value for lcd brightness: %d", vconf_get_ext_errno());
		ret = -EPERM;
		goto error;
	}

	ret = vconf_get_int(VCONFKEY_SETAPPL_BRIGHTNESS_AUTOMATIC_INT, &autobrt);
	if (ret < 0) {
		_E("Failed to get vconf value for automatic brightness: %d", vconf_get_ext_errno());
		ret = -EPERM;
		goto error;
	}

	ret = vconf_set_int(VCONFKEY_PM_CUSTOM_BRIGHTNESS_STATUS, VCONFKEY_PM_CUSTOM_BRIGHTNESS_OFF);
	if (ret < 0)
		_E("Failed to set vconf value for custom brightness status: %d", vconf_get_ext_errno());

	ret = backlight_ops.get_brightness(&brt);
	if (ret < 0)
		brt = ret;

	// check dim state
	if (low_battery_state(bat) &&
	    charger == VCONFKEY_SYSMAN_CHARGER_DISCONNECTED && !changed) {
		_D("batt warning low : brightness is not changed!");
		if (brt != 0)
			backlight_ops.set_brightness(0);
		goto error;
	}

	if (autobrt == SETTING_BRIGHTNESS_AUTOMATIC_OFF) {
		if (brt != setting)
			backlight_ops.set_brightness(setting);
	} else if (autobrt == SETTING_BRIGHTNESS_AUTOMATIC_PAUSE) {
		_D("Auto brightness will be enable");
		ret = vconf_set_int(VCONFKEY_SETAPPL_BRIGHTNESS_AUTOMATIC_INT, SETTING_BRIGHTNESS_AUTOMATIC_ON);
		if (ret < 0)
			_E("Failed to set vconf value for automatic brightness: %d", vconf_get_ext_errno());
	}

error:
	return g_variant_new("(i)", ret);
}

GVariant *dbus_setrefreshrate(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int app, val, ret, control;

	g_variant_get(param, "(ii)", &app, &val);

	if (app < 0 || app >= ARRAY_SIZE(display_conf.framerate_app) || val < 0) {
		ret = -EINVAL;
		goto error;
	}

	if (!display_conf.framerate_app[app]) {
		_I("This case(%d) is not support in this target", app);
		ret = -EPERM;
		goto error;
	}

	control = display_conf.control_display;

	if (control)
		backlight_ops.off(NORMAL_MODE);

	_D("app : %d, value : %d", app, val);
	ret = backlight_ops.set_frame_rate(val);
	if (ret < 0)
		_E("Failed to set frame rate (%d)", ret);

	if (control)
		backlight_ops.on(NORMAL_MODE);

error:
	return g_variant_new("(i)", ret);
}

GVariant *dbus_setautobrightnessmin(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int val, ret;
	pid_t pid;
	int id = 0;

	if (!display_info.set_autobrightness_min) {
		ret = -EIO;
		goto error;
	}
	g_variant_get(param, "(i)", &val);

	pid = dbus_connection_get_sender_pid(conn, sender);
	ret = display_info.set_autobrightness_min(val, (char *)sender);
	if (ret) {
		_W("fail to set autobrightness min %d, %d by %d", val, ret, pid);
		goto error;
	}
	if (display_info.reset_autobrightness_min) {
		id = dbus_handle_watch_name(sender, display_info.reset_autobrightness_min, NULL, NULL, NULL);
		if (id <= 0) {
			_E("failed to watch name %s, id %d", sender, id);
			//todo: set return value
		}
		_I("set autobrightness min %d by %d", val, pid);
	}
error:
	return g_variant_new("(i)", ret);
}

GVariant *dbus_setlcdtimeout(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int on, dim, holdkey_block, ret;
	pid_t pid;
	int id = 0;

	g_variant_get(param, "(iii)", &on, &dim, &holdkey_block);

	pid = dbus_connection_get_sender_pid(conn, sender);
	ret = set_lcd_timeout(on, dim, holdkey_block, sender);
	if (ret) {
		_W("fail to set lcd timeout %d by %d", ret, pid);
	} else {
		id = dbus_handle_watch_name(sender, reset_lcd_timeout, NULL, NULL, NULL);
		if (id <= 0) {
			_E("failed to watch name %s, id %d", sender, id);
			//todo: set return value
		}
		_I("set lcd timeout on %d, dim %d, holdblock %d by %d",
		    on, dim, holdkey_block, pid);
	}

	return g_variant_new("(i)", ret);
}

GVariant *dbus_lockscreenbgon(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret = 0;
	char *on = NULL;

	g_variant_get(param, "(s)", &on);

	if (!strcmp(on, "true")) {
		if (disp_plgn.update_pm_setting)
			disp_plgn.update_pm_setting(SETTING_LOCK_SCREEN_BG, true);
	} else if (!strcmp(on, "false")) {
		if (disp_plgn.update_pm_setting)
			disp_plgn.update_pm_setting(SETTING_LOCK_SCREEN_BG, false);
	} else
		ret = -EINVAL;

	g_free(on);
	return g_variant_new("(i)", ret);
}

GVariant *dbus_dumpmode(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret = 0;
	char *on;

	g_variant_get(param, "(s)", &on);

	if (!strcmp(on, "on")) {
		pm_save_logdump();
		if (disp_plgn.pm_lock_internal)
			disp_plgn.pm_lock_internal(INTERNAL_LOCK_DUMPMODE, LCD_OFF,
				STAY_CUR_STATE, DUMP_MODE_WATING_TIME);
		backlight_ops.release_blink();
	} else if (!strcmp(on, "off")) {
		if (disp_plgn.pm_unlock_internal)
			disp_plgn.pm_unlock_internal(INTERNAL_LOCK_DUMPMODE, LCD_OFF, PM_SLEEP_MARGIN);
		backlight_ops.blink(0);
	} else if (!strcmp(on, "blink")) {
		backlight_ops.blink(500);
	} else {
		ret = -EINVAL;
	}

	g_free(on);
	return g_variant_new("(i)", ret);
}

GVariant *dbus_savelog(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	pm_save_logdump();
	return dbus_handle_new_g_variant_tuple();
}

GVariant *dbus_powerkeyignore(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret = 0;
	int on;

	g_variant_get(param, "(i)", &on);

	if (CHECK_OPS(keyfilter_ops, set_powerkey_ignore))
		keyfilter_ops->set_powerkey_ignore(on == 1 ? true : false);

	return g_variant_new("(i)", ret);
}

GVariant *dbus_powerkeylcdoff(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret;

	if (CHECK_OPS(keyfilter_ops, powerkey_lcdoff))
		ret = keyfilter_ops->powerkey_lcdoff();
	else
		ret = -ENOSYS;

	return g_variant_new("(i)", ret);
}

GVariant *dbus_customlcdon(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret = 0;
	int timeout;

	g_variant_get(param, "(i)", &timeout);

	ret = custom_lcdon(timeout);

	return g_variant_new("(i)", ret);
}

GVariant *dbus_customlcdoff(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret = 0;
	enum device_flags flag;
	char *reason_str;

	g_variant_get(param, "(s)", &reason_str);

	if (!strcmp(reason_str, LCDOFF_PROXI_STR))
		flag = LCD_OFF_BY_PROXIMITY;
	else if (!strcmp(reason_str, LCDOFF_GESTURE_STR))
		flag = LCD_OFF_BY_GESTURE;
	else {
		_E("%s resean is invalid, dbus ignored!", reason_str);
		ret = -EINVAL;
		goto out;
	}

	ret = custom_lcdoff(flag);

out:
	g_free(reason_str);

	return g_variant_new("(i)", ret);
}

GVariant *dbus_changestatebyreason(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret = 0;
	int next_state, timeout;
	char *reason;

	g_variant_get(param, "(isi)", &next_state, &reason, &timeout);

	if (next_state == DISPLAY_STATE_NORMAL)
		ret = display_on_by_reason(reason, timeout);
	else if (next_state == DISPLAY_STATE_SCREEN_OFF)
		ret = display_off_by_reason(reason);
	else
		ret = -EINVAL;

	g_free(reason);
	return g_variant_new("(i)", ret);
}

GVariant *dbus_staytouchscreenoff(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret = 0;
	int val;

	g_variant_get(param, "(i)", &val);

	set_stay_touchscreen_off(val);

	return g_variant_new("(i)", ret);
}

GVariant *dbus_lcdpaneloffmode(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret = 0;
	int val;

	g_variant_get(param, "(i)", &val);

	set_lcd_paneloff_mode(val);

	return g_variant_new("(i)", ret);
}

GVariant *dbus_actorcontrol(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret = 0, val, actor;
	char *op;

	g_variant_get(param, "(sii)", &op, &actor, &val);

	if (!strcmp(op, "set"))
		ret = display_set_caps(actor, val);
	else if (!strcmp(op, "reset"))
		ret = display_reset_caps(actor, val);
	else
		ret = -EINVAL;

	g_free(op);
	return g_variant_new("(i)", ret);
}

GVariant *dbus_getcustombrightness(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int status = 0;

	status = backlight_ops.get_custom_status();

	return g_variant_new("(i)", status);
}

struct pmlock_expired_s {
	char req_id[32];
	GDBusMethodInvocation *invocation;
	int value;
	pid_t pid;
};

static dd_list *expired_req_list;

static int get_command(pid_t pid, char *comm, size_t len)
{
	FILE *fp;
	char path[PATH_MAX];
	char name[NAME_MAX];
	char *tmp;
	int ret;

	snprintf(path, sizeof(path), "/proc/%d/comm", pid);

	fp = fopen(path, "r");
	if (!fp) {
		_E("Failed to open file %s, errno:%d", path, errno);
		return -errno;
	}

	ret = fread(name, 1, sizeof(name) - 1, fp);
	name[ret] = '\0';
	fclose(fp);

	tmp = name;
	while (*tmp != '\0') {
		if (*tmp == '\n' ||
				*tmp == '\r') {
			*tmp = '\0';
			break;
		}
		tmp++;
	}

	snprintf(comm, len, "%s", name);

	return 0;
}

GVariant *dbus_locktimeout_expired(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret;
	char *req_id;
	pid_t pid;
	char pid_str[16];
	char comm[NAME_MAX];
	struct pmlock_expired_s *ex = NULL;
	GVariant *gvar = NULL;

	g_variant_get(param, "(s)", &req_id);

	pid = dbus_connection_get_sender_pid(conn, sender);
	if (kill(pid, 0) == -1) {
		_E("%d process does not exist, dbus ignored!", pid);
		ret = -ESRCH;
		goto out;
	}

	ret = get_command(pid, comm, sizeof(comm));
	if (ret < 0) {
		_E("Failed to get command (%d)", ret);
		goto out;
	}

	ex = calloc(1, sizeof(struct pmlock_expired_s));
	if (!ex) {
		_E("calloc() failed");
		ret = -ENOMEM;
		goto out;
	}

	snprintf(pid_str, sizeof(pid_str), "%d", pid);
	ret = launch_system_app(APP_DEFAULT, 8,
			APP_KEY_TYPE,
			EXPIRED_POPUP_TYPE_POWER,
			EXPIRED_POPUP_PID,
			pid_str,
			EXPIRED_POPUP_COMM,
			comm,
			EXPIRED_POPUP_ID,
			req_id);
	if (ret < 0) {
		_E("Failed to launch pmlock expired popup(%d)", ret);
		goto out;
	}

	ex->pid = pid;
	snprintf(ex->req_id, sizeof(ex->req_id), "%s", req_id);

	ex->invocation = invocation;

	DD_LIST_APPEND(expired_req_list, ex);

	g_free(req_id);
	return NULL;

out:
	gvar = g_variant_new("(i)", ret);
	g_free(req_id);
	if (ex)
		free(ex);
	return gvar;
}

static gboolean app_term(gpointer data)
{
	struct pmlock_expired_s *ex = data;

	if (ex) {
		if (ex->pid > 0 && kill(ex->pid, 0) != -1) {
			if (kill(ex->pid, SIGKILL) == -1)
				_E("failed to kill pid(%d), err(%m)", ex->pid);
		}
		free(ex);
	}
	return G_SOURCE_REMOVE;
}

static void expired_deliver_result(void *data)
{
	struct pmlock_expired_s *ex = data;
	struct pmlock_expired_s *item = NULL;
	dd_list *l;
	size_t len;
	int value;
	char *id;

	if (!ex)
		return;

	len = strlen(ex->req_id) + 1;
	DD_LIST_FOREACH(expired_req_list, l, item) {
		if (!strncmp(item->req_id, ex->req_id, len)) {
			DD_LIST_REMOVE(expired_req_list, item);
			break;
		}
	}
	if (!item)
		goto out;

	id = ex->req_id;
	value = ex->value;

	_I("User input of req_id(%s) is (%s)", id, value == 0 ? "ALLOW power lock" :
			(value == 2) ? "KILL app" : "Release power lock");

	if (!item->invocation) {
		_E("item->invocation is null");
		goto out;
	}
	g_dbus_method_invocation_return_value(item->invocation, g_variant_new("(i)", 0));

	if (value == 2) {
		if (item->pid > 0 && kill(item->pid, SIGTERM) == -1)
			_E("failed to kill pid(%d), err(%m)", item->pid);
		g_timeout_add(3000, app_term, (gpointer)item);
		item = NULL;
	}

out:
	free(ex);
	free(item);
}

GVariant *dbus_locktimeout_input(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret;
	char *req_id;
	struct pmlock_expired_s *ex;
	int value;

	g_variant_get(param, "(si)", &req_id, &value);

	ex = calloc(1, sizeof(struct pmlock_expired_s));
	if (!ex) {
		_E("calloc() failed");
		ret = -ENOMEM;
		goto out;
	}

	snprintf(ex->req_id, sizeof(ex->req_id), "%s", req_id);
	ex->value = value;

	_I("req_id(%s), value(%d)", ex->req_id, ex->value);

	add_idle_request(expired_deliver_result, ex);

	ret = 0;

out:
	g_free(req_id);
	return g_variant_new("(i)", ret);
}

static GVariant *dbus_dimstay_control(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int dimstay;

	g_variant_get(param, "(i)", &dimstay);

	if (dimstay < 0) {
		_E("Invalid dimstay set value %d.", dimstay);
		return g_variant_new("(i)", -EINVAL);
	}

	if (dimstay) {
		_I("Set DIM_FLAG.");
		pm_status_flag |= DIM_FLAG;
	} else {
		_I("Unset DIM_FLAG.");
		pm_status_flag &= ~DIM_FLAG;
	}

	if (pm_cur_state == S_NORMAL)
		backlight_ops.update();

	return g_variant_new("(i)", 0);
}

static const dbus_method_s dbus_methods[] = {
	{ "start",           NULL,  NULL, dbus_start },
	{ "stop",            NULL,  NULL, dbus_stop },
	{ "lockstate",     "sssi",   "i", dbus_lockstate },
	{ "unlockstate",     "ss",   "i", dbus_unlockstate },
	{ "changestate",      "s",   "i", dbus_changestate },
	{ "ChangeState",      "s",   "i", dbus_changestate },
	{ "getbrightness",    "i",   "i", dbus_getbrightness },	/* deprecated */
	{ "setbrightness",   "ii",   "i", dbus_setbrightness },	/* deprecated */
	{ "setframerate",    "ii",   "i", dbus_setrefreshrate },	/* deprecated */
	{ "setautobrightnessmin", "i", "i", dbus_setautobrightnessmin },
	{ "setlcdtimeout",  "iii",   "i", dbus_setlcdtimeout },
	{ "LockScreenBgOn",   "s",   "i", dbus_lockscreenbgon },
	{ "GetDisplayCount", NULL,   "i", dbus_getdisplaycount },
	{ "GetMaxBrightness", "i",  "i", dbus_getmaxbrightness },
	{ "SetMaxBrightness", "i",   "i", dbus_setmaxbrightness },
	{ "GetBrightness",    "i",   "i", dbus_getbrightness },
	{ "SetBrightness",   "ii",   "i", dbus_setbrightness },
	{ "HoldBrightness",   "i",   "i", dbus_holdbrightness },
	{ "ReleaseBrightness", NULL, "i", dbus_releasebrightness },
	{ "SetRefreshRate",  "ii",   "i", dbus_setrefreshrate },
	{ "Dumpmode",         "s",   "i", dbus_dumpmode },
	{ "SaveLog",         NULL,  NULL, dbus_savelog },
	{ "PowerKeyIgnore",   "i",   "i", dbus_powerkeyignore },
	{ "PowerKeyLCDOff",  NULL,   "i", dbus_powerkeylcdoff },
	{ "CustomLCDOn",      "i",   "i", dbus_customlcdon },
	{ "CustomLCDOff",	  "s",	 "i", dbus_customlcdoff },
	{ "ChangeStateByReason", "isi", "i", dbus_changestatebyreason },
	{ "StayTouchScreenOff", "i", "i", dbus_staytouchscreenoff },
	{ "LCDPanelOffMode",  "i",   "i", dbus_lcdpaneloffmode },
	{ "ActorControl",   "sii",   "i", dbus_actorcontrol },
	{ "CustomBrightness", NULL,  "i", dbus_getcustombrightness },
	{ "CurrentBrightness", "i", "i", dbus_getbrightness }, /* deprecated. It is remained for tizen 2.4 */
	{ "LockTimeoutExpired", "s", "i", dbus_locktimeout_expired },
	{ "LockTimeoutInput",  "si", "i", dbus_locktimeout_input },
	{ "DimStayControl",        "i",  "i", dbus_dimstay_control },
	/* Add methods here */
};

static const dbus_interface_u dbus_interface = {
	.oh = NULL,
	.name = DEVICED_INTERFACE_DISPLAY,
	.methods = dbus_methods,
	.nr_methods = ARRAY_SIZE(dbus_methods),
};

static void homescreen_signal_handler(GDBusConnection  *conn,
										const gchar      *sender,
										const gchar      *path,
										const gchar      *iface,
										const gchar      *name,
										GVariant         *param,
										gpointer          data)
{
	char *screen;
	pid_t pid;

	g_variant_get(param, "(s)", &screen);

	_D("screen : %s", screen);

	pid = dbus_connection_get_sender_pid(conn, sender);
	ambient_set_screen(screen, pid);
	g_free(screen);
}

static void sim_signal_handler(GDBusConnection  *conn,
										const gchar      *sender,
										const gchar      *path,
										const gchar      *iface,
										const gchar      *name,
										GVariant         *param,
										gpointer          data)
{
	int ret, val;
	static int state = false;

	ret = vconf_get_bool(VCONFKEY_LCD_BRIGHTNESS_INIT, &state);
	if (ret < 0 || state) {
		_E("Failed to get %s", VCONFKEY_LCD_BRIGHTNESS_INIT);
		return;
	}

	g_variant_get(param, "(i)", &val);

	if (val != SIM_CARD_NOT_PRESENT) {
		state = true;
		ret = vconf_set_bool(VCONFKEY_LCD_BRIGHTNESS_INIT, state);
		if (ret < 0)
				_E("Failed to set vconf value for lcd brightness: %d", vconf_get_ext_errno());
		ret = vconf_set_int(VCONFKEY_SETAPPL_LCD_BRIGHTNESS, PM_DEFAULT_BRIGHTNESS);
		if (ret < 0)
				_E("Failed to set vconf value for lcd brightness: %d", vconf_get_ext_errno());
		backlight_ops.set_brightness(PM_DEFAULT_BRIGHTNESS);
		_I("SIM card is inserted at first.");
	}
}

static void changestate_signal_handler(GDBusConnection  *conn,
										const gchar      *sender,
										const gchar      *path,
										const gchar      *iface,
										const gchar      *name,
										GVariant         *param,
										gpointer          data)
{
	int val;
	char *appid, *pkgid, *state, *type;
	pid_t pid;

	g_variant_get(param, "(issss)", &val, &appid, &pkgid, &state, &type);

	pid = (pid_t)val;

	if (!strcmp(state, "bg")) {
		_D("process(%d) was going background.", pid);
		device_notify(DEVICE_NOTIFIER_PROCESS_BACKGROUND, &pid);
	} else if (!strcmp(state, "fg")) {
		_D("process(%d) was going foreground.", pid);
		device_notify(DEVICE_NOTIFIER_PROCESS_FOREGROUND, &pid);
	}
	g_free(appid);
	g_free(pkgid);
	g_free(state);
	g_free(type);
}

/*
 * Default capability
 * api      := LCDON | LCDOFF | BRIGHTNESS
 * gesture  := LCDON
 */
static struct display_actor_ops display_api_actor = {
	.id	= DISPLAY_ACTOR_API,
	.caps	= DISPLAY_CAPA_LCDON |
		  DISPLAY_CAPA_LCDOFF |
		  DISPLAY_CAPA_BRIGHTNESS,
};

static struct display_actor_ops display_gesture_actor = {
	.id	= DISPLAY_ACTOR_GESTURE,
	.caps	= DISPLAY_CAPA_LCDON,
};

int init_pm_dbus(void)
{
	int ret;

	display_add_actor(&display_api_actor);
	display_add_actor(&display_gesture_actor);

	ret = dbus_handle_add_dbus_object(NULL, DEVICED_PATH_DISPLAY, &dbus_interface);
	if (ret < 0)
		_E("fail to init dbus method(%d)", ret);

#ifndef MICRO_DD
	ret = subscribe_dbus_signal(NULL,
				TELEPHONY_PATH,
				TELEPHONY_INTERFACE_SIM,
				SIGNAL_SIM_STATUS,
				sim_signal_handler,
				NULL, NULL);
	if (ret <= 0)
		_E("Failed to register signal handler: %d", ret);
#endif

	ret = subscribe_dbus_signal(NULL,
				DEVICED_OBJECT_PATH,
				DEVICED_INTERFACE_NAME,
				SIGNAL_HOMESCREEN,
				homescreen_signal_handler,
				NULL, NULL);
	if (ret <= 0)
		_E("Failed to register signal handler: %d", ret);

	ret = subscribe_dbus_signal(NULL,
				AUL_APPSTATUS_PATH,
				AUL_APPSTATUS_INTERFACE,
				APP_CHANGE_STATE,
				changestate_signal_handler,
				NULL, NULL);
	if (ret <= 0)
		_E("Failed to register signal handler: %d", ret);

	return 0;
}

