/*
 * deviced
 *
 * Copyright (c) 2016 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <stdio.h>
#include <stdbool.h>
#include <vconf.h>
#include <bundle.h>
#include <eventsystem.h>

#include "core/log.h"
#include "core/list.h"
#include "core/common.h"
#include "core/device-notifier.h"
#include "display/poll.h"
#include "extcon/extcon.h"
#include "apps/apps.h"
#include "usb.h"
#include "usb-tethering.h"
#include "usb-debug.h"

#define MTP_RESPONDER_PATH	"/usr/lib/systemd/system/mtp-responder.service"

static int noti_id = -1;
static unsigned int usb_current_mode = USB_FUNCTION_NONE;
static unsigned int usb_selected_mode = USB_FUNCTION_SDB; /* USB_FUNCTION_SDB is for debugging */
static extcon_usb_state_e usb_connection = USB_DISCONNECTED;

/**
 * Important
 * You must keep the order for the specific items.
 *
 * There may be several mode_vconfs for a mode.
 * In this case, the representative value should be placed on top.
 *
 * In case of SET_USB_SDB_DIAG and SET_USB_DIAG_SDB, SET_USB_SDB_DIAG should be above.
 * As another example, SEt_USB_RNDIS_SDB should be at the top of them.
 */
static const struct _usb_mode_mapping_table {
	int mode_vconf;
	unsigned int mode;
} usb_mode_mapping_table[] = {
	/* Hack for performance. In most cases this is the value. */
	{SET_USB_DEFAULT,         USB_FUNCTION_MTP | USB_FUNCTION_ACM},
	{SET_USB_SDB,             USB_FUNCTION_MTP | USB_FUNCTION_ACM | USB_FUNCTION_SDB},

	{SET_USB_NONE,            USB_FUNCTION_NONE},
	{SET_USB_RNDIS,           USB_FUNCTION_RNDIS},
	{SET_USB_RNDIS_DIAG,      USB_FUNCTION_RNDIS | USB_FUNCTION_DIAG},
	{11,                      USB_FUNCTION_DIAG | USB_FUNCTION_RMNET}, /* SET_USB_DIAG_RMNET */
	{12,                      USB_FUNCTION_ACM | USB_FUNCTION_SDB | USB_FUNCTION_DM}, /* SET_USB_ACM_SDB_DM */

	{SET_USB_SDB_DIAG,        USB_FUNCTION_MTP | USB_FUNCTION_ACM | USB_FUNCTION_SDB | USB_FUNCTION_DIAG},
	{SET_USB_DIAG_SDB,        USB_FUNCTION_MTP | USB_FUNCTION_ACM | USB_FUNCTION_SDB | USB_FUNCTION_DIAG},

	{SET_USB_RNDIS_SDB,       USB_FUNCTION_ACM | USB_FUNCTION_SDB | USB_FUNCTION_RNDIS},
	{SET_USB_RNDIS_TETHERING, USB_FUNCTION_ACM | USB_FUNCTION_SDB | USB_FUNCTION_RNDIS},
	{13,                      USB_FUNCTION_ACM | USB_FUNCTION_SDB | USB_FUNCTION_RNDIS}, /* SET_USB_RNDIS_SDB_ACM */
};

static unsigned int selected_mode_from_vconf(int mode_vconf)
{
	int i;
	int array_len = sizeof(usb_mode_mapping_table)/sizeof(usb_mode_mapping_table[0]);

	for (i = 0; i < array_len; i++) {
		if (usb_mode_mapping_table[i].mode_vconf == mode_vconf)
			return usb_mode_mapping_table[i].mode;
	}

	return USB_FUNCTION_NONE;
}

static int selected_mode_to_vconf(unsigned int mode)
{
	int i;
	int array_len = sizeof(usb_mode_mapping_table)/sizeof(usb_mode_mapping_table[0]);

	/* Caution: The order to search the usb_mode_mapping_table must start with a low index. */
	for (i = 0; i < array_len; i++) {
		if (usb_mode_mapping_table[i].mode == mode)
			return usb_mode_mapping_table[i].mode_vconf;
	}

	return SET_USB_DEFAULT; /* To prevent usb abnormal operation, use it instead of SET_USB_NONE */
}

static int current_mode_to_vconf(unsigned int mode)
{
	bool mtp = mode & USB_FUNCTION_MTP;
	bool sdb = mode & USB_FUNCTION_SDB;
	bool acm = mode & USB_FUNCTION_ACM;
	bool diag = mode & USB_FUNCTION_DIAG;
	bool rndis = mode & USB_FUNCTION_RNDIS;
	bool dm = mode & USB_FUNCTION_DM;
	bool rmnet = mode & USB_FUNCTION_RMNET;

	if (mode == USB_FUNCTION_NONE)
		return SET_USB_NONE;

	if (mtp && acm) {
		if (sdb) {
			if (diag)
				return SET_USB_SDB_DIAG;
			return SET_USB_SDB;
		}
		return SET_USB_DEFAULT;
	}

	if (rndis)
		return SET_USB_RNDIS;

	if (diag && sdb)
		return SET_USB_DIAG_SDB;

	if (diag && rmnet)
		return 11; /* SET_USB_DIAG_RMNET */

	if (acm && sdb && dm)
		return 12; /* SET_USB_ACM_SDB_DM */

	return -EINVAL;
}

static void usb_state_send_system_event(int status)
{
	bundle *b;
	const char *str;

	switch (status) {
	case VCONFKEY_SYSMAN_USB_DISCONNECTED:
		str = EVT_VAL_USB_DISCONNECTED;
		break;
	case VCONFKEY_SYSMAN_USB_CONNECTED:
		str = EVT_VAL_USB_CONNECTED;
		break;
	case VCONFKEY_SYSMAN_USB_AVAILABLE:
		str = EVT_VAL_USB_AVAILABLE;
		break;
	default:
		return;
	}

	_I("system_event(%s)", str);

	b = bundle_create();
	bundle_add_str(b, EVT_KEY_USB_STATUS, str);
	eventsystem_send_system_event(SYS_EVENT_USB_STATUS, b);
	bundle_free(b);
}

static void usb_state_set_connection(extcon_usb_state_e conn)
{
	usb_connection = conn;
}

extcon_usb_state_e usb_state_get_connection(void)
{
	return usb_connection;
}

void usb_state_retrieve_selected_mode(void)
{
	int ret;
	int mode_vconf;

	ret = vconf_get_int(VCONFKEY_USB_SEL_MODE, &mode_vconf);
	if (ret < 0) {
		_E("Failed to get vconf value for usb sel mode: %d", vconf_get_ext_errno());
		return;
	}

	usb_selected_mode = selected_mode_from_vconf(mode_vconf);
	_I("usb_select_mode_vconf=%d, condition=%x", mode_vconf, usb_selected_mode);

#ifdef ENGINEER_MODE
	_I("Engineer mode. usb selected mode %d, debug state %d", usb_selected_mode, get_usb_debug_state());

	if (!(usb_selected_mode & USB_FUNCTION_SDB)) {
		usb_selected_mode = USB_FUNCTION_MTP | USB_FUNCTION_ACM | USB_FUNCTION_SDB;
		(void)usb_state_set_selected_mode(usb_selected_mode);
		set_usb_debug_state(true);
	}
#endif

	if (access(MTP_RESPONDER_PATH, R_OK) != 0) {
		usb_selected_mode &= ~USB_FUNCTION_MTP;
		(void)usb_state_set_selected_mode(usb_selected_mode);
	}
}

/* Cache of vconf_get_int(VCONFKEY_USB_SEL_MODE) */
unsigned int usb_state_get_selected_mode(void)
{
	return usb_selected_mode;
}

int usb_state_set_selected_mode(unsigned int mode)
{
	int ret;
	int mode_vconf;

	usb_selected_mode = mode;

	mode_vconf = selected_mode_to_vconf(mode);
	ret = vconf_set_int(VCONFKEY_USB_SEL_MODE, mode_vconf);
	if (ret < 0)
		_E("Failed to set vconf value for usb selected mode: %d", vconf_get_ext_errno());

	return ret;
}

/* Cache of vconf_get_int(VCONFKEY_USB_CUR_MODE) */
unsigned int usb_state_get_current_mode(void)
{
	return usb_current_mode;
}

int usb_state_set_current_mode(unsigned int mode)
{
	int ret = 0;
	int mode_vconf;
	static int old_mode_vconf = -1;

	usb_current_mode = mode;

	mode_vconf = current_mode_to_vconf(mode);
	if (mode_vconf >= 0 && old_mode_vconf != mode_vconf) {
		old_mode_vconf = mode_vconf;
		ret = vconf_set_int(VCONFKEY_USB_CUR_MODE, mode_vconf);
		if (ret < 0)
			_E("Failed to set vconf value for usb current mode: %d", vconf_get_ext_errno());
	}

	return ret;
}

#if 0
char *usb_state_get_mode_str(unsigned int mode, char *str, size_t len)
{
	int i;
	int ret;
	int flag = 0;

	if (mode == USB_FUNCTION_NONE) {
		snprintf(str, len, "%s", "");
	} else {
		for (i = 0; i < ARRAY_SIZE(_available_funcs); ++i) {
			if (mode & _available_funcs[i]->id) {
				ret = snprintf(str, len, "%s,",
					       _available_funcs[i]->name);
				if (ret >= len)
					return str;
				str += ret;
				len -= ret;
				flag = 1;
			}
		}
		if (flag == 1) {
			/* eliminate trailing comma */
			*(str - 1) = '\0';
		}
	}

	return str;
}
#endif

static void media_noti_cb(GVariant *var, void *user_data, GError *err)
{
	int id = 0;

	if (!var) {
		if (err)
			_E("Error: %s", err->message);
		return;
	}

	if (!dh_get_param_from_var(var, "(i)", &id)) {
		_E("Failed to get variant(%s): no message", g_variant_get_type_string(var));
		goto out;
	}

	noti_id = id;
	_D("Media noti(%d)", noti_id);

out:
	g_variant_unref(var);
}

void usb_state_update_state(extcon_usb_state_e state, unsigned int mode)
{
	int ret;
	int status;
	static int old_mode = -1;
	static int old_status = -1; /* VCONFKEY_SYSMAN_USB_STATUS */

	if (state == USB_DISCONNECTED && mode != USB_FUNCTION_NONE)
		mode = USB_FUNCTION_NONE;

	if (mode == USB_FUNCTION_NONE) {
		if (noti_id >= 0) {
			int ret = remove_notification("MediaDeviceNotiOff", noti_id);
			if (ret < 0)
				_E("Failed to remove event_noti: %d", ret);
			else
				noti_id = -1;
		}
	} else if (mode & USB_FUNCTION_MTP) {
		if (noti_id < 0) {
			ret = add_async_notification("MediaDeviceNotiOn", media_noti_cb, NULL, NULL);
			if (ret < 0)
				_E("Failed to show notification for usb connection.");
		}
	}

	usb_state_set_connection(state);
	if (state == USB_CONNECTED) {
		if (mode == USB_FUNCTION_NONE)
			status = VCONFKEY_SYSMAN_USB_CONNECTED;
		else
			status = VCONFKEY_SYSMAN_USB_AVAILABLE;
	} else
		status = VCONFKEY_SYSMAN_USB_DISCONNECTED;

	if (old_status != status) {
		usb_state_send_system_event(status);
		ret = vconf_set_int(VCONFKEY_SYSMAN_USB_STATUS, status);
		if (ret < 0)
			_E("Failed to set vconf value for usb status: %d", vconf_get_ext_errno());
		broadcast_usb_state_changed();
		old_status = status;
	}

	if (old_mode != mode) {
		(void)usb_state_set_current_mode(mode);
		old_mode = mode;
	}
}
